<?php

/**
 * Waiting Queue Aggregate Queue class.
 */
final class WaitingQueueAggregateQueue implements DrupalQueueInterface, WaitingQueueBlockingQueueInterface {

  protected $subQueues;

  /**
   * @inheritdoc
   */
  public function createItem($data) {
    throw new WaitingQueueAggregateQueuecreateItemException('Adding items to this queue is forbidden.');
  }

  /**
   * Wrap a subqueue item.
   *
   * @param string $queue
   *   The subqueue name.
   * @param mixed $item
   *   The subqueue item.
   *
   * @return object
   *   The wrapped subqueue item.
   */
  protected function wrapItem($queue, $item) {
    $this->statsRecordItemProcessed();
    return (object) array(
      'data' => array(
        'queue' => $queue,
        'item' => $item,
      ),
    );
  }

  /**
   * Unwrap a queued sub item.
   *
   * @param mixed $item
   *   The queue item.
   *
   * @return array
   *   An array containing the original queue item and the queue name.
   */
  protected function unwrapItem($item) {
    return array(
      'queue' => $item->data['queue'],
      'item' => $item->data['item'],
    );
  }

  protected static $queueCallbacks = array();

  /**
   * Get the worker callback for the configured subqueue.
   *
   * @param string $name
   *   The name of the queue.
   *
   * @return callable
   *   The queue callback for this queue.
   */
  protected static function getQueueCallback($name) {
    if (!isset(self::$queueCallbacks[$name])) {
      self::$queueCallbacks[$name] = waiting_queue_get_callback_function($name);
    }
    return self::$queueCallbacks[$name];
  }

  /**
   * Get an array of all configured subqueues.
   *
   * @return DrupalQueueInterface[]
   *   The array of configured subqueues.
   */
  protected function getSubqueues() {
    if (!isset($this->subQueues)) {
      $this->subQueues = array();
      $enabled_queues = variable_get('waiting_queue_aggregate_enabled_queues', array());
      $cron_queues = module_invoke_all('cron_queue_info');
      drupal_alter('cron_queue_info', $cron_queues);
      foreach ($cron_queues as $queue_name => $subqueue_info) {
        if (in_array($queue_name, $enabled_queues, TRUE) && isset($cron_queues[$queue_name]['worker callback'])) {
          $this->subQueues[$queue_name] = DrupalQueue::get($queue_name);
        }
      }

      // Check for modules implementing hook_queue_info().
      $queues = module_invoke_all('queue_info');
      drupal_alter('queue_info', $queues);
      foreach ($queues as $queue_name => $subqueue_info) {
        if (in_array($queue_name, $enabled_queues, TRUE) && isset($cron_queues[$queue_name]['callback'])) {
          $this->subQueues[$queue_name] = DrupalQueue::get($queue_name);
        }
      }
    }
    return $this->subQueues;
  }

  /**
   * Get a given subqueue.
   *
   * @param string $name
   *   The name of the queue to get the queue class for.
   *
   * @return DrupalQueueInterface
   *   The queue.
   */
  protected function getSubqueue($name) {
    $queues = $this->getSubqueues();
    return isset($queues[$name]) ? $queues[$name] : NULL;
  }

  /**
   * @inheritdoc
   */
  public function numberOfItems() {
    $items = 0;
    foreach ($this->getSubqueues() as $queue) {
      $items += $queue->numberOfItems();
    }
    return $items;
  }

  /**
   * @inheritdoc
   */
  public function createQueue() {
    // We do not store items directly, so nothing to do here.
  }

  /**
   * @inheritdoc
   */
  public function deleteQueue() {
    // We do not store items directly, so nothing to do here.
  }

  /**
   * @inheritdoc
   */
  public function deleteItem($item) {
    $unwrapped = $this->unwrapItem($item);
    $queue = $this->getSubqueue($unwrapped['queue']);
    $queue->deleteItem($unwrapped['item']);
  }

  /**
   * @inheritdoc
   */
  public function releaseItem($item) {
    $unwrapped = $this->unwrapItem($item);
    $queue = $this->getSubqueue($unwrapped['queue']);
    $queue->releaseItem($unwrapped['item']);
  }

  /**
   * Try and claim an item from the subqueues.
   *
   * @param int $lease_time
   *   The time to claim an item for.
   *
   * @return mixed
   *   FALSE if no item could be claimed, or an item if it could.
   *
   * @throws WaitingQueueAggregateQueueclaimItemFromSubqueuesException
   */
  public function claimItem($lease_time = 900) {
    $this->ensureStartStats();
    // If we're not blocking, just pass through to the original implementation.
    if (!$this->blocking) {
      return $this->claimItemFromSubqueues($lease_time);
    }
    // Otherwise, start a polling loop.
    do {
      // Try to get an item from the queue.
      $item = $this->claimItemFromSubqueues($lease_time);
      // If we didn't get an item, then usleep for the polling interval, and
      // repeat the loop. We need the || TRUE at the end, because the usleep
      // function returns NULL which would exit our loop.
    } while (($item === FALSE) && (usleep($this->pollingInterval) || TRUE));
    return $item;
  }

  /**
   * A list of sub queue names this class is working through.
   *
   * @var string[]
   */
  protected $workQueues;

  /**
   * A list of sub queues this class currently considers empty.
   *
   * @var string[]
   */
  protected $noWorkQueues;

  /**
   * Claim a single item from one of the configured subqueues.
   *
   * @param int $lease_time
   *   The time to claim an item for.
   *
   * @return bool|object
   *   Either FALSE if no item could be claimed, or the claimed item.
   *
   * @throws WaitingQueueAggregateQueueclaimItemFromSubqueuesException
   */
  protected function claimItemFromSubqueues($lease_time = 900) {
    // Setup the queues to work through.
    if (empty($this->workQueues)) {
      // Reset to having no queues without items.
      $this->noWorkQueues = array();

      // Create a list of sub queue names, duplicated 25 times over.
      foreach (array_keys($this->getSubqueues()) as $subqueue_name) {
        for ($i = 0; $i < 25; $i++) {
          $this->workQueues[] = $subqueue_name;
        }
      }

      // Shuffle the possible work into random order.
      shuffle($this->workQueues);

      if (empty($this->workQueues)) {
        throw new WaitingQueueAggregateQueueclaimItemFromSubqueuesException('No subqueues configured for processing.');
      }
    }

    while ($subqueue_name = array_pop($this->workQueues)) {
      if (!in_array($subqueue_name, $this->noWorkQueues, TRUE)) {
        if ($item = $this->getSubqueue($subqueue_name)
          ->claimItem($lease_time)) {
          // Return a sub queue item, wrapped to include the sub queue name.
          return $this->wrapItem($subqueue_name, $item);
        }
        else {
          // No work in this sub queue for the moment.
          $this->noWorkQueues[] = $subqueue_name;
        }
      }
    }

    return FALSE;
  }

  /**
   * Boolean indicating if this queue's claimItem method is going to block.
   *
   * @var bool
   */
  protected $blocking = FALSE;

  /**
   * The the amount of time in microseconds to wait between polling the queue.
   *
   * @var int
   */
  public $pollingInterval = 1000000;

  /**
   * Change the blocking behavior of the claimItem method.
   *
   * By default the claimItem method in this class will not block until there
   * is an item to return, but optionally you can use this method to change
   * this behavior so that it will block.
   *
   * @param mixed $blocking
   *   The new blocking status to set.
   */
  public function setClaimItemBlocking($blocking) {
    $this->blocking = $blocking;
  }

  protected $startTime;
  protected $itemsProcessed = 0;
  protected $processID;

  protected function ensureStartStats() {
    if (is_null($this->startTime)) {
      $this->startTime = time();
      $this->processID = uniqid(mt_rand(), TRUE);
    }
  }

  protected function writeStats($isTerminating = FALSE) {
    if (!is_null($this->startTime)) {
      db_merge('waiting_queue_aggregate_stats')
        ->key(array(
          'process_id' => $this->processID,
          ))
        ->fields(array(
          'start_time' => $this->startTime,
          'items_processed' => $this->itemsProcessed,
          'last_seen_time' => time(),
          'status' => $isTerminating ? t('Terminated') : t('Processing'),
        ))
        ->execute();
    }
  }

  public function __destruct() {
    $this->writeStats(TRUE);
  }

  protected function statsRecordItemProcessed() {
    $this->itemsProcessed++;

    if ($this->itemsProcessed > 0 && $this->itemsProcessed % 10 == 1) {
      $this->writeStats();
    }
  }


  /**
   * The worker callback.
   *
   * @param mixed $item
   *   The wrapped queue item to process.
   *
   * @see waiting_queue_aggregate_worker_callback
   *
   * @throws WaitingQueueAggregateQueueprocessItemException
   */
  public static function workerCallback($item) {
    $queue_callback = self::getQueueCallback($item['queue']);
    if (!empty($queue_callback)) {
      call_user_func($queue_callback, $item['item']->data);
    }
    else {
      throw new WaitingQueueAggregateQueueprocessItemException('Could not find queue callback for: ' . $item['queue']);
    }
  }

}
